import * as A from "@automerge/automerge";
import debug from "debug";
import EventEmitter from "eventemitter3";
import { assign, createMachine, interpret, } from "xstate";
import { waitFor } from "xstate/lib/waitFor.js";
import { headsAreSame } from "./helpers/headsAreSame.js";
import { pause } from "./helpers/pause.js";
import { withTimeout, TimeoutError } from "./helpers/withTimeout";
/** DocHandle is a wrapper around a single Automerge document that lets us listen for changes. */
export class DocHandle//
 extends EventEmitter {
    documentId;
    #log;
    #machine;
    #timeoutDelay;
    constructor(documentId, { isNew = false, timeoutDelay = 700000 } = {}) {
        super();
        this.documentId = documentId;
        this.#timeoutDelay = timeoutDelay;
        this.#log = debug(`automerge-repo:dochandle:${documentId.slice(0, 5)}`);
        // initial doc
        const doc = A.init({
            patchCallback: (patches, { before, after }) => this.emit("patch", { handle: this, patches, before, after }),
        });
        /**
         * Internally we use a state machine to orchestrate document loading and/or syncing, in order to
         * avoid requesting data we already have, or surfacing intermediate values to the consumer.
         *
         *                      ┌─────────┐           ┌────────────┐
         *  ┌───────┐  ┌──FIND──┤ loading ├─REQUEST──►│ requesting ├─UPDATE──┐
         *  │ idle  ├──┤        └───┬─────┘           └────────────┘         │
         *  └───────┘  │            │                                        └─►┌─────────┐
         *             │            └───────LOAD───────────────────────────────►│  ready  │
         *             └──CREATE───────────────────────────────────────────────►└─────────┘
         */
        this.#machine = interpret(createMachine({
            predictableActionArguments: true,
            id: "docHandle",
            initial: IDLE,
            context: { documentId, doc },
            states: {
                idle: {
                    on: {
                        // If we're creating a new document, we don't need to load anything
                        CREATE: { target: READY },
                        // If we're accessing an existing document, we need to request it from storage
                        // and/or the network
                        FIND: { target: LOADING },
                        DELETE: { actions: "onDelete", target: DELETED },
                    },
                },
                loading: {
                    on: {
                        // LOAD is called by the Repo if the document is found in storage
                        LOAD: { actions: "onLoad", target: READY },
                        // REQUEST is called by the Repo if the document is not found in storage
                        REQUEST: { target: REQUESTING },
                        DELETE: { actions: "onDelete", target: DELETED },
                    },
                },
                requesting: {
                    on: {
                        // UPDATE is called by the Repo when we receive changes from the network
                        UPDATE: { actions: "onUpdate" },
                        // REQUEST_COMPLETE is called from `onUpdate` when the doc has been fully loaded from the network
                        REQUEST_COMPLETE: { target: READY },
                        DELETE: { actions: "onDelete", target: DELETED },
                    },
                },
                ready: {
                    on: {
                        // UPDATE is called by the Repo when we receive changes from the network
                        UPDATE: { actions: "onUpdate", target: READY },
                        DELETE: { actions: "onDelete", target: DELETED },
                    },
                },
                error: {},
                deleted: {},
            },
        }, {
            actions: {
                /** Apply the binary changes from storage and put the updated doc on context */
                onLoad: assign((context, { payload }) => {
                    const { binary } = payload;
                    const { doc } = context;
                    const workaroundDoc = A.load(binary, { unchecked: true });
                    const secondBinary = A.save(workaroundDoc);
                    const newDoc = A.loadIncremental(doc, secondBinary);
                    return { doc: newDoc };
                }),
                /** Put the updated doc on context; if it's different, emit a `change` event */
                onUpdate: assign((context, { payload }) => {
                    const { doc: oldDoc } = context;
                    const { callback } = payload;
                    const newDoc = callback(oldDoc);
                    const docChanged = !headsAreSame(newDoc, oldDoc);
                    if (docChanged) {
                        this.emit("change", { handle: this, doc: newDoc });
                        if (!this.isReady()) {
                            this.#machine.send(REQUEST_COMPLETE);
                        }
                    }
                    return { doc: newDoc };
                }),
                onDelete: assign(() => {
                    this.emit("delete", { handle: this });
                    return { doc: undefined };
                }),
            },
        }))
            .onTransition(({ value: state }, { type: event }) => this.#log(`${event} → ${state}`, this.#doc))
            .start();
        this.#machine.send(isNew ? CREATE : FIND);
    }
    get doc() {
        if (!this.isReady()) {
            throw new Error(`DocHandle#${this.documentId} is not ready. Check \`handle.isReady()\` before accessing the document.`);
        }
        return this.#doc;
    }
    // PRIVATE
    /** Returns the current document */
    get #doc() {
        return this.#machine?.getSnapshot().context.doc;
    }
    /** Returns the docHandle's state (READY, etc.) */
    get #state() {
        return this.#machine?.getSnapshot().value;
    }
    /** Returns a promise that resolves when the docHandle is in one of the given states */
    #statePromise(awaitStates) {
        if (!Array.isArray(awaitStates))
            awaitStates = [awaitStates];
        return Promise.any(awaitStates.map(state => waitFor(this.#machine, s => s.matches(state))));
    }
    // PUBLIC
    isReady = () => this.#state === READY;
    isDeleted = () => this.#state === DELETED;
    /**
     * Returns the current document, waiting for the handle to be ready if necessary.
     */
    async value(awaitStates = [READY]) {
        await pause(); // yield one tick because reasons
        try {
            // wait for the document to enter one of the desired states
            await withTimeout(this.#statePromise(awaitStates), this.#timeoutDelay);
        }
        catch (error) {
            if (error instanceof TimeoutError)
                throw new Error(`DocHandle: timed out loading ${this.documentId}`);
            else
                throw error;
        }
        // Return the document
        return this.#doc;
    }
    async loadAttemptedValue() {
        return this.value([READY, REQUESTING]);
    }
    /** `load` is called by the repo when the document is found in storage */
    load(binary) {
        if (binary.length) {
            this.#machine.send(LOAD, { payload: { binary } });
        }
    }
    /** `update` is called by the repo when we receive changes from the network */
    update(callback) {
        this.#machine.send(UPDATE, { payload: { callback } });
    }
    /** `change` is called by the repo when the document is changed locally  */
    change(callback, options = {}) {
        if (!this.isReady()) {
            throw new Error(`DocHandle#${this.documentId} is not ready. Check \`handle.isReady()\` before accessing the document.`);
        }
        this.#machine.send(UPDATE, {
            payload: {
                callback: (doc) => {
                    return A.change(doc, options, callback);
                },
            },
        });
    }
    /** `request` is called by the repo when the document is not found in storage */
    request() {
        if (this.#state === LOADING)
            this.#machine.send(REQUEST);
    }
    /** `delete` is called by the repo when the document is deleted */
    delete() {
        this.#machine.send(DELETE);
    }
}
// STATE MACHINE TYPES
// state
export const HandleState = {
    IDLE: "idle",
    LOADING: "loading",
    REQUESTING: "requesting",
    READY: "ready",
    ERROR: "error",
    DELETED: "deleted",
};
// events
export const Event = {
    CREATE: "CREATE",
    LOAD: "LOAD",
    FIND: "FIND",
    REQUEST: "REQUEST",
    REQUEST_COMPLETE: "REQUEST_COMPLETE",
    UPDATE: "UPDATE",
    TIMEOUT: "TIMEOUT",
    DELETE: "DELETE",
};
// CONSTANTS
const { IDLE, LOADING, REQUESTING, READY, ERROR, DELETED } = HandleState;
const { CREATE, LOAD, FIND, REQUEST, UPDATE, TIMEOUT, DELETE, REQUEST_COMPLETE, } = Event;
